def test_rebind_symbols(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    hilda_client.rebind_symbols()
    assert len(hilda_client.symbols) > 0



import logging

from construct import Array, CString, Hex, If, Int32ub, Int32ul, Int64ul, Pointer, Struct, Tell, this
from humanfriendly import prompts

from hilda.lldb_importer import lldb
from hilda.snippets.macho.image_info import ImageInfo, dyld_image_info_t
from hilda.snippets.macho.macho import mach_header_t
from hilda.snippets.uuid import uuid_t
from hilda.symbol import SymbolFormatField

all_image_infos_t = Struct(
    'address' / Tell,
    'version' / Int32ul,
    'infoArrayCount' / Int32ul,
    '_infoArray' / SymbolFormatField(lldb.hilda_client),
    'infoArray' / Pointer(this._infoArray, Array(this.infoArrayCount, dyld_image_info_t)),
    'notification' / SymbolFormatField(lldb.hilda_client),
    'processDetachedFromSharedRegion' / Int32ub,
    'libSystemInitialized' / Int32ub,
    '_dyldImageLoadAddress' / SymbolFormatField(lldb.hilda_client),
    'dyldImageLoadAddress' / Pointer(this._dyldImageLoadAddress, mach_header_t),
    'jitInfo' / SymbolFormatField(lldb.hilda_client),
    '_dyldVersion' / SymbolFormatField(lldb.hilda_client),
    'dyldVersion' / Pointer(this._dyldVersion, CString('utf8')),
    '_errorMessage' / SymbolFormatField(lldb.hilda_client),
    'errorMessage' / If(this._errorMessage != 0, Pointer(this._erroressage, CString('utf8'))),
    'terminationFlags' / Int64ul,
    'coreSymbolicationShmPage' / SymbolFormatField(lldb.hilda_client),
    'systemOrderFlags' / Int64ul,
    'uuidArrayCount' / Int64ul,
    '_uuidArray' / SymbolFormatField(lldb.hilda_client),
    'uuidArray' / If(this._uuidArray != 0,
                     Pointer(this._uuidArray, Array(this.uuidArrayCount, uuid_t))),
    '_dyld_all_image_infos' / SymbolFormatField(lldb.hilda_client),
    # 'dyld_all_image_infos' / Lazy(LazyBound(lambda: Pointer(this._dyld_all_image_infos, self.all_image_infos))),
    'initialImageCount' / Int64ul,
    'errorKind' / Int64ul,
    '_errorClientOfDyLibPath' / SymbolFormatField(lldb.hilda_client),
    'errorClientOfDyLibPath' / If(this._errorClientOfDyLibPath != 0,
                                  Pointer(this._errorClientofoyLibPath, CString('utf8'))),
    '_errorTargetDylibPath' / SymbolFormatField(lldb.hilda_client),
    'errorTargetDylibPath' / If(this._errorTargetDylibPath != 0, Pointer(this._errorTargetDylibPath, CString('utf8'))),
    '_errorSymbol' / SymbolFormatField(lldb.hilda_client),
    'errorSymbol' / If(this._errorSymbol != 0, Pointer(this._errorSymbol, CString('utf8'))),
    'sharedCacheSlide' / Hex(Int64ul),
)


class AllImageInfos(object):
    def reload(self):
        with lldb.hilda_client.stopped(1):
            all_image_infos_symbol = lldb.hilda_client.symbol(lldb.hilda_client.symbols.dyld_all_image_infos)
            self.__all_image_infos = all_image_infos_t.parse_stream(all_image_infos_symbol)
            self.__image_infos = []

            for image_info_data in self.__all_image_infos.infoArray:
                image_info = ImageInfo(image_info_data)
                self.__image_infos.append(image_info)

    def __init__(self):
        self.__all_image_infos = None
        self.__image_infos = None

        self.reload()

    @property
    def images(self):
        return self.__image_infos

    def find_images(self, name):
        return [image for image in self.images if name in image.file_path]

    def image_dependencies(self, image_name):
        images = [image for image in self.find_images(image_name)]

        if len(images) == 0:
            # This is a time consuming action. We only do it in
            # because a library could be dynamically opened
            self.reload()
            images = [image for image in self.find_images(image_name)]

            if len(images) == 0:
                print(f"Image {image_name} is not loaded")
                return

        return AllImageInfos.__image_dependencies(images)

    @staticmethod
    def __image_dependencies(images):
        unique_images = []

        for image in images:
            if image.file_path not in [unique_image.file_path for unique_image in unique_images]:
                unique_images.append(image)

        image = unique_images[0]
        if len(unique_images) > 1:
            image = AllImageInfos.__select_specific_image(unique_images)

        dependencies = set([image_name.path for image_name in image.load_commands.dylib_commands])

        return dependencies

    @staticmethod
    def __select_specific_image(images):
        logging.info('Multiple images were found with that prefix.\nPlease select one')
        print([f"{image.file_path}" for image in images])
        selected_image_path = prompts.prompt_for_choice([f"{image.file_path}" for image in images])
        for image in images:
            if image.file_path == selected_image_path:
                return image



import pytest

from hilda.exceptions import LLDBException
from hilda.hilda_client import HildaClient
from hilda.launch_lldb import LLDBAttachName
from hilda.lldb_importer import lldb

PROCESS = 'sysmond'


@pytest.fixture(scope='session')
def lldb_debugger(request):
    lldb_t = None
    try:
        lldb_t = LLDBAttachName(PROCESS)
        lldb_t.start()
        yield lldb_t.debugger
    except LLDBException as e:
        pytest.exit(f'{e.message} - Try use sudo')
    finally:
        if lldb_t:
            lldb_t.process.Detach()
            lldb_t.join()


@pytest.fixture(scope='function')
def hilda_client(lldb_debugger):
    client = HildaClient(lldb_debugger)
    lldb.hilda_client = client
    client.init_dynamic_environment()
    with client.stopped():
        yield client



from hilda.snippets import xpc


def test_from_xpc_object(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    xpc_object = hilda_client.evaluate_expression('''
        typedef void * xpc_object_t;
        typedef void * xpc_activity_t;
        typedef void (^xpc_activity_handler_t)(xpc_activity_t activity);
        extern xpc_object_t xpc_dictionary_create(const char * const *keys, const xpc_object_t *values, size_t);
        extern void xpc_activity_register(const char *, xpc_object_t criteria, xpc_activity_handler_t handler);
        extern void xpc_dictionary_set_int64(xpc_object_t xdict, const char *key, int64_t value);
        extern void xpc_dictionary_set_string(xpc_object_t xdict, const char *key, const char *string);

        xpc_object_t criteria = xpc_dictionary_create(NULL, NULL, 0);
        xpc_dictionary_set_int64(criteria, "Delay", 5);
        xpc_dictionary_set_int64(criteria, "GracePeriod", 1);
        xpc_dictionary_set_string(criteria, "Priority", "Utility");
        criteria;
    ''')
    assert xpc.from_xpc_object(xpc_object) == {'Delay': 5, 'Priority': 'Utility', 'GracePeriod': 1}



import pytest

from hilda.objective_c_class import PropertyAttributes, convert_encoded_property_attributes


def test_calling_class_method_by_method_name(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    NSDictionary = hilda_client.objc_get_class('NSDictionary')
    dictionary = NSDictionary.dictionaryWithObjects_forKeys_count_(0, 0, 0)
    assert dictionary.po() == '{\n}'
    dictionary = NSDictionary['dictionaryWithObjects:forKeys:count:'](0, 0, 0)
    assert dictionary.po() == '{\n}'


def test_calling_class_method_via_objc_call(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    NSDictionary = hilda_client.objc_get_class('NSDictionary')
    dictionary = NSDictionary.objc_call('dictionaryWithObjects:forKeys:count:', 0, 0, 0)
    assert dictionary.po() == '{\n}'


def test_getting_class_method(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    NSDictionary = hilda_client.objc_get_class('NSDictionary')
    assert 'dictionaryWithObjects_forKeys_count_' in dir(NSDictionary)


def test_calling_super_class_method(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    NSDictionary = hilda_client.objc_get_class('NSDictionary')
    dictionary = NSDictionary.new()
    assert dictionary.po() == '{\n}'


def test_getting_super_class_method(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    NSDictionary = hilda_client.objc_get_class('NSDictionary')
    assert 'new' in dir(NSDictionary)


@pytest.mark.parametrize('encoded, result', [
    ('Tc,VcharDefault', PropertyAttributes(synthesize='charDefault', type_='char', list=[])),
    ('T{YorkshireTeaStruct=ic},VstructDefault',
     PropertyAttributes(synthesize='structDefault', type_='struct YorkshireTeaStruct { int x0; char x1; }', list=[])),
    ('T@,R,&,VidReadonlyRetainNonatomic',
     PropertyAttributes(synthesize='idReadonlyRetainNonatomic', type_='id', list=['readonly', 'strong'])),
    ('T^v', PropertyAttributes(synthesize=None, type_='void *', list=[])),
])
def test_convert_encoded_property_attributes(encoded: str, result: PropertyAttributes):
    """
    :param encoded: Property encoding.
    :param result: Parsed property data.
    """
    assert convert_encoded_property_attributes(encoded) == result


@pytest.mark.parametrize('sub_str', [
    '@interface NSObject',
    '+ alloc;',
    '@property (readonly,copy) NSString * description;',
])
def test_class_without_super_str(hilda_client, sub_str: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param sub_str: Substring that should appear in the str.
    """
    assert str(hilda_client.objc_get_class('NSObject')).count(sub_str) == 1



from prompt_toolkit.enums import DEFAULT_BUFFER
from prompt_toolkit.filters import EmacsInsertMode, HasFocus, HasSelection, ViInsertMode
from prompt_toolkit.keys import Keys


def load_ipython_extension(ipython):
    def register_keybindings():
        hilda = ipython.user_ns['p']
        keys_mapping = {Keys.F7: hilda.step_into,
                        Keys.F8: hilda.step_over,
                        Keys.F9: lambda _: (hilda.log_info('Sending continue'), hilda.cont()),
                        Keys.F10: lambda _: (hilda.log_info('Sending stop'), hilda.stop())}

        insert_mode = ViInsertMode() | EmacsInsertMode()
        registry = ipython.pt_app.key_bindings

        for key, callback in keys_mapping.items():
            registry.add_binding(key, filter=(HasFocus(DEFAULT_BUFFER) & ~HasSelection() & insert_mode))(
                callback)

    register_keybindings()
    ipython.events.register('shell_initialized', register_keybindings)



import base64
import builtins
import importlib
import importlib.util
import json
import logging
import os
import pickle
import struct
import sys
import time
import typing
from collections import namedtuple
from contextlib import contextmanager, suppress
from dataclasses import dataclass, field
from datetime import datetime, timezone
from functools import cached_property, wraps
from pathlib import Path
from typing import Any, Callable, List, Optional, Union

import hexdump
import IPython
from humanfriendly import prompts
from humanfriendly.terminal.html import html_to_ansi
from IPython.core.magic import register_line_magic  # noqa: F401
from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import XmlLexer
from tqdm import tqdm
from traitlets.config import Config

from hilda import objective_c_class
from hilda.common import CfSerializable
from hilda.exceptions import AccessingMemoryError, AccessingRegisterError, AddingLldbSymbolError, \
    BrokenLocalSymbolsJarError, ConvertingFromNSObjectError, ConvertingToNsObjectError, CreatingObjectiveCSymbolError, \
    DisableJetsamMemoryChecksError, EvaluatingExpressionError, HildaException, SymbolAbsentError
from hilda.lldb_importer import lldb
from hilda.objective_c_symbol import ObjectiveCSymbol
from hilda.registers import Registers
from hilda.snippets.mach import CFRunLoopServiceMachPort_hooks
from hilda.symbol import Symbol
from hilda.symbols_jar import SymbolsJar
from hilda.ui.ui_manager import UiManager

lldb.KEYSTONE_SUPPORT = True
try:
    from keystone import KS_ARCH_ARM64, KS_ARCH_X86, KS_MODE_64, KS_MODE_LITTLE_ENDIAN, Ks
except ImportError:
    lldb.KEYSTONE_SUPPORT = False
    print('failed to import keystone. disabling some features')

hilda_art = Path(__file__).resolve().parent.joinpath('hilda_ascii_art.html').read_text()

GREETING = f"""
{hilda_art}

<b>Hilda has been successfully loaded! 😎
Usage:
 <span style="color: magenta">p</span>   Global to access all features.
 <span style="color: magenta">F7</span>  Step Into.
 <span style="color: magenta">F8</span>  Step Over.
 <span style="color: magenta">F9</span>  Continue.
 <span style="color: magenta">F10</span> Stop.

Have a nice flight ✈️! Starting an IPython shell...
"""


def disable_logs() -> None:
    logging.getLogger('asyncio').disabled = True
    logging.getLogger('parso.cache').disabled = True
    logging.getLogger('parso.cache.pickle').disabled = True
    logging.getLogger('parso.python.diff').disabled = True
    logging.getLogger('humanfriendly.prompts').disabled = True
    logging.getLogger('blib2to3.pgen2.driver').disabled = True
    logging.getLogger('hilda.launch_lldb').disabled = True


SerializableSymbol = namedtuple('SerializableSymbol', 'address type_ filename')


@dataclass
class Configs:
    """ Configuration settings for evaluation and monitoring. """
    evaluation_unwind_on_error: bool = field(default=False,
                                             metadata={'doc': 'Whether to unwind on error during evaluation.'})
    evaluation_ignore_breakpoints: bool = field(default=False,
                                                metadata={'doc': 'Whether to ignore breakpoints during evaluation.'})
    nsobject_exclusion: bool = field(default=False, metadata={
        'doc': 'Whether to exclude NSObject during evaluation - reduce ipython autocomplete results.'})
    objc_verbose_monitor: bool = field(default=False, metadata={
        'doc': 'When set to True, using monitor() will automatically print objc methods arguments.'})

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        config_str = 'Configuration settings:\n'
        max_len = max(len(field_name) for field_name in self.__dataclass_fields__) + 2

        for field_name, field_info in self.__dataclass_fields__.items():
            value = getattr(self, field_name)
            doc = field_info.metadata.get('doc', 'No docstring available')
            config_str += f'\t{field_name.ljust(max_len)}: {str(value).ljust(5)} | {doc}\n'

        return config_str


def stop_is_needed(func: Callable):
    """Decorator to check if the process must be stopped before proceeding."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        is_running = self.process.GetState() == lldb.eStateRunning
        if is_running:
            self.logger.error(f'Cannot {func.__name__.replace("_", "-")}: Process must be stopped first.')
            return
        return func(self, *args, **kwargs)

    return wrapper


class HildaClient:
    Breakpoint = namedtuple('Breakpoint', 'address options forced callback')

    RETVAL_BIT_COUNT = 64

    def __init__(self, debugger: lldb.SBDebugger):
        self.logger = logging.getLogger(__name__)
        self.endianness = '<'
        self.debugger = debugger
        self.target = debugger.GetSelectedTarget()
        self.process = self.target.GetProcess()
        self.symbols = SymbolsJar.create(self)
        self.breakpoints = {}
        self.captured_objects = {}
        self.registers = Registers(self)
        self.arch = self.target.GetTriple().split('-')[0]
        self.ui_manager = UiManager(self)
        self.configs = Configs()
        self._dynamic_env_loaded = False
        self._symbols_loaded = False

        # the frame called within the context of the hit BP
        self._bp_frame = None

        self._add_global('symbols', self.symbols, [])
        self._add_global('registers', self.registers, [])

        self.log_info(f'Target: {self.target}')
        self.log_info(f'Process: {self.process}')

    def hd(self, buf):
        """
        Print an hexdump of given buffer
        :param buf: buffer to print in hexdump form
        """
        print(hexdump.hexdump(buf))

    def lsof(self) -> dict:
        """
        Get dictionary of all open FDs
        :return: Mapping between open FDs and their paths
        """
        data = (Path(__file__).parent / 'objective_c' / 'lsof.m').read_text()
        result = json.loads(self.po(data))
        # convert FDs into int
        return {int(k): v for k, v in result.items()}

    def bt(self, should_print: bool = False, depth: Optional[int] = None) -> List[Union[str, lldb.SBFrame]]:
        """ Print an improved backtrace. """
        backtrace = []
        for i, frame in enumerate(self.thread.frames):
            if i == depth:
                break
            row = ''
            row += html_to_ansi(f'<span style="color: cyan">0x{frame.addr.GetFileAddress():x}</span> ')
            row += str(frame)
            if i == 0:
                # first line
                row += ' 👈'
            backtrace.append([f'0x{frame.addr.file_addr:016x}', frame])
            if should_print:
                print(row)
        return backtrace

    def disable_jetsam_memory_checks(self):
        """
        Disable jetsam memory checks, prevent raising:
        `error: Execution was interrupted, reason: EXC_RESOURCE RESOURCE_TYPE_MEMORY (limit=15 MB, unused=0x0).`
        when evaluating expression.
        """
        # 6 is for MEMORYSTATUS_CMD_SET_JETSAM_TASK_LIMIT
        result = self.symbols.memorystatus_control(6, self.process.GetProcessID(), 0, 0, 0)
        if result:
            raise DisableJetsamMemoryChecksError()

    def symbol(self, address):
        """
        Get symbol object for a given address
        :param address:
        :return: Hilda's symbol object
        """
        return Symbol.create(address, self)

    def objc_symbol(self, address) -> ObjectiveCSymbol:
        """
        Get objc symbol wrapper for given address
        :param address:
        :return: Hilda's objc symbol object
        """
        try:
            return ObjectiveCSymbol.create(int(address), self)
        except HildaException as e:
            raise CreatingObjectiveCSymbolError from e

    def inject(self, filename):
        """
        Inject a single library into currently running process
        :param filename:
        :return: module object
        """
        module = self.target.FindModule(lldb.SBFileSpec(os.path.basename(filename), False))
        if module.file.basename is not None:
            self.log_warning(f'file {filename} has already been loaded')

        injected = SymbolsJar.create(self)
        handle = self.symbols.dlopen(filename, 10)  # RTLD_GLOBAL|RTLD_NOW

        if handle == 0:
            self.log_critical(f'failed to inject: {filename}')

        module = self.target.FindModule(lldb.SBFileSpec(os.path.basename(filename), False))
        for symbol in module.symbols:
            load_addr = symbol.addr.GetLoadAddress(self.target)
            if load_addr == 0xffffffffffffffff:
                # skip those not having a real address
                continue

            name = symbol.name
            type_ = symbol.GetType()

            if name in ('<redacted>',) or (type_ not in (lldb.eSymbolTypeCode,
                                                         lldb.eSymbolTypeData,
                                                         lldb.eSymbolTypeObjCMetaClass)):
                # ignore unnamed symbols and those which are not: data, code or objc classes
                continue

            injected[name] = self.symbol(load_addr)
        return injected

    def rebind_symbols(self, image_range=None, filename_expr=''):
        """
        Reparse all loaded images symbols
        :param image_range: index range for images to load in the form of [start, end]
        :param filename_expr: filter only images containing given expression
        """
        self.log_debug('mapping symbols')
        self._symbols_loaded = False

        for i, module in enumerate(tqdm(self.target.modules)):
            filename = module.file.basename

            if filename_expr not in filename:
                continue

            if image_range is not None and (i < image_range[0] or i > image_range[1]):
                continue

            for symbol in module:
                with suppress(AddingLldbSymbolError):
                    self.add_lldb_symbol(symbol)

        globals()['symbols'] = self.symbols
        self._symbols_loaded = True

    @stop_is_needed
    def poke(self, address, buf: bytes):
        """
        Write data at given address
        :param address:
        :param buf:
        """
        err = lldb.SBError()
        retval = self.process.WriteMemory(address, buf, err)

        if not err.Success():
            raise AccessingMemoryError()

        return retval

    @stop_is_needed
    def poke_text(self, address: int, code: str) -> int:
        """
        Write instructions to address.
        :param address:
        :param code:
        """
        if not lldb.KEYSTONE_SUPPORT:
            raise NotImplementedError('Not supported without keystone')
        bytecode, count = self._ks.asm(code, as_bytes=True)
        return self.poke(address, bytecode)

    @stop_is_needed
    def peek(self, address, size: int) -> bytes:
        """
        Read data at given address
        :param address:
        :param size:
        :return:
        """
        if size == 0:
            return b''

        err = lldb.SBError()
        retval = self.process.ReadMemory(address, int(size), err)

        if not err.Success():
            raise AccessingMemoryError()

        return retval

    @stop_is_needed
    def peek_str(self, address: Symbol) -> str:
        """
        Peek a buffer till null termination
        :param address:
        :return:
        """
        return address.po('char *')[1:-1]  # strip the ""

    def stop(self, *args) -> None:
        """ Stop process. """
        self.debugger.SetAsync(False)

        is_running = self.process.GetState() == lldb.eStateRunning
        if not is_running:
            self.log_debug('already stopped')
            return

        if not self.process.Stop().Success():
            self.log_critical('failed to stop process')

    def cont(self, *args) -> None:
        """ Continue process. """
        is_running = self.process.GetState() == lldb.eStateRunning

        if is_running:
            self.log_debug('already running')
            return

        # bugfix:   the debugger may become in sync state, so we make sure
        #           it isn't before trying to continue
        self.debugger.SetAsync(True)

        if not self.process.Continue().Success():
            self.log_critical('failed to continue process')

    def detach(self):
        """
        Detach from process.

        Useful in order to exit gracefully so process doesn't get killed
        while you exit
        """
        if not self.process.Detach().Success():
            self.log_critical('failed to detach')
        else:
            self.log_info('Process Detached')

    @stop_is_needed
    def disass(self, address: int, buf: bytes, flavor: str = 'intel',
               should_print: bool = False) -> lldb.SBInstructionList:
        """
        Print disassembly from a given address
        :param flavor:
        :param address:
        :param buf:
        :param should_print:
        :return:
        """
        inst = self.target.GetInstructionsWithFlavor(lldb.SBAddress(address, self.target), flavor, buf)
        if should_print:
            print(inst)
        return inst

    def file_symbol(self, address: int, module_name: Optional[str] = None) -> Symbol:
        """
        Calculate symbol address without ASLR
        :param address: address as can be seen originally in Mach-O
        :param module_name: Module name to resolve the symbol from
        """
        if module_name is None:
            module = self.target
        else:
            module = self.target.FindModule(lldb.SBFileSpec(module_name))

        return self.symbol(module.ResolveFileAddress(address).GetLoadAddress(self.target))

    def get_register(self, name) -> Symbol:
        """
        Get value for register by its name
        :param name:
        :return:
        """
        register = self.frame.register[name.lower()]
        if register is None:
            raise AccessingRegisterError()
        return self.symbol(register.unsigned)

    def set_register(self, name, value):
        """
        Set value for register by its name
        :param name:
        :param value:
        :return:
        """
        register = self.frame.register[name.lower()]
        if register is None:
            raise AccessingRegisterError()
        register.value = hex(value)

    def objc_call(self, obj, selector, *params):
        """
        Simulate a call to an objc selector
        :param obj: obj to pass into `objc_msgSend`
        :param selector: selector to execute
        :param params: any other additional parameters the selector requires
        :return: invocation returned value
        """
        # On object `obj`
        args = self._serialize_call_params([obj])
        # Call selector (by its uid)
        args.append(self._generate_call_expression(self.symbols.sel_getUid, self._serialize_call_params([selector])))
        # With params
        args.extend(self._serialize_call_params(params))
        call_expression = self._generate_call_expression(self.symbols.objc_msgSend, args)
        with self.stopped():
            return self.evaluate_expression(call_expression)

    def call(self, address, argv: list = None):
        """
        Call function at given address with given parameters
        :param address:
        :param argv: parameter list
        :return: function's return value
        """
        if argv is None:
            argv = []
        call_expression = self._generate_call_expression(address, self._serialize_call_params(argv))
        with self.stopped():
            return self.evaluate_expression(call_expression)

    def monitor(self, address, condition: str = None, **options) -> lldb.SBBreakpoint:
        """
        Monitor every time a given address is called

        The following options are available:
            regs={reg1: format}
                will print register values

                Available formats:
                    x: hex
                    s: string
                    cf: use CFCopyDescription() to get more informative description of the object
                    po: use LLDB po command
                    std::string: for std::string

                    User defined function, will be called like `format_function(hilda_client, value)`.

                For example:
                    regs={'x0': 'x'} -> x0 will be printed in HEX format
            expr={lldb_expression: format}
                lldb_expression can be for example '$x0' or '$arg1'
                format behaves just like 'regs' option
            retval=format
                Print function's return value. The format is the same as regs format.
            stop=True
                force a stop at every hit
            bt=True
                print backtrace
            cmd=[cmd1, cmd2]
                run several LLDB commands, one by another
            force_return=value
                force a return from function with the specified value
            name=some_value
                use `some_name` instead of the symbol name automatically extracted from the calling frame
            override=True
                override previous break point at same location


        :param address:
        :param condition: set as a conditional breakpoint using an lldb expression
        :param options:
        :return:
        """

        def callback(hilda, frame, bp_loc, options):
            """
            :param HildaClient hilda: Hilda client.
            :param lldb.SBFrame frame: LLDB Frame object.
            :param lldb.SBBreakpointLocation bp_loc: LLDB Breakpoint location object.
            :param dict options: User defined options.
            """
            bp = bp_loc.GetBreakpoint()
            symbol = hilda.breakpoints[bp.id].address  # type: Symbol

            # by default, attempt to resolve the symbol name through lldb
            name = str(symbol.lldb_symbol)
            if options.get('name', False):
                name = options['name']

            log_message = f'🚨 #{bp.id} 0x{symbol:x} {name}'

            if 'regs' in options:
                log_message += '\nregs:'
                for name, fmt in options['regs'].items():
                    value = hilda.symbol(frame.FindRegister(name).unsigned)
                    log_message += f'\n\t{name} = {hilda._monitor_format_value(fmt, value)}'

            if 'expr' in options:
                log_message += '\nexpr:'
                for name, fmt in options['expr'].items():
                    value = hilda.symbol(hilda.evaluate_expression(name))
                    log_message += f'\n\t{name} = {hilda._monitor_format_value(fmt, value)}'

            if options.get('force_return', False):
                hilda.force_return(options['force_return'])
                log_message += f'\nforced return: {options["force_return"]}'

            if options.get('bt', False):
                # bugfix: for callstacks from xpc events
                hilda.finish()
                hilda.bt()

            if options.get('retval', False):
                # return from function
                hilda.finish()
                value = hilda.evaluate_expression('$arg1')
                log_message += f'\nreturned: {hilda._monitor_format_value(options["retval"], value)}'

            hilda.log_info(log_message)

            for cmd in options.get('cmd', []):
                hilda.lldb_handle_command(cmd)

            if not options.get('stop', False):
                hilda.cont()

        return self.bp(address, callback, condition=condition, **options)

    def show_current_source(self):
        """ print current source code if possible """
        self.lldb_handle_command('f')

    def finish(self):
        """ Run current frame till its end. """
        with self.sync_mode():
            self.thread.StepOutOfFrame(self.frame)
            self._bp_frame = None

    @stop_is_needed
    def step_into(self, *args):
        """ Step into current instruction. """
        with self.sync_mode():
            self.thread.StepInto()
        if self.ui_manager.active:
            self.ui_manager.show()

    @stop_is_needed
    def step_over(self, *args):
        """ Step over current instruction. """
        with self.sync_mode():
            self.thread.StepOver()
        if self.ui_manager.active:
            self.ui_manager.show()

    def remove_all_hilda_breakpoints(self, remove_forced=False):
        """
        Remove all breakpoints created by Hilda
        :param remove_forced: include removed of "forced" breakpoints
        """
        breakpoints = list(self.breakpoints.items())
        for bp_id, bp in breakpoints:
            if remove_forced or not bp.forced:
                self.remove_hilda_breakpoint(bp_id)

    def remove_hilda_breakpoint(self, bp_id):
        """
        Remove a single breakpoint placed by Hilda
        :param bp_id: Breakpoint's ID
        """
        self.target.BreakpointDelete(bp_id)
        del self.breakpoints[bp_id]
        self.log_info(f'BP #{bp_id} has been removed')

    def force_return(self, value=0):
        """
        Prematurely return from a stack frame, short-circuiting exection of newer frames and optionally
        yielding a specified value.
        :param value:
        :return:
        """
        self.finish()
        self.set_register('x0', value)

    def proc_info(self):
        """ Print information about currently running mapped process. """
        print(self.process)

    def print_proc_entitlements(self):
        """ Get the plist embedded inside the process' __LINKEDIT section. """
        linkedit_section = self.target.modules[0].FindSection('__LINKEDIT')
        linkedit_data = self.symbol(linkedit_section.GetLoadAddress(self.target)).peek(linkedit_section.size)

        # just look for the xml start inside the __LINKEDIT section. should be good enough since wer'e not
        # expecting any other XML there
        entitlements = str(linkedit_data[linkedit_data.find(b'<?xml'):].split(b'\xfa', 1)[0], 'utf8')
        print(highlight(entitlements, XmlLexer(), TerminalTrueColorFormatter()))

    def bp(self, address_or_name: Union[int, str], callback: Optional[Callable] = None, condition: str = None,
           forced=False, module_name: Optional[str] = None, **options) -> lldb.SBBreakpoint:
        """
        Add a breakpoint
        :param address_or_name:
        :param condition: set as a conditional breakpoint using lldb expression
        :param callback: callback(hilda, *args) to be called
        :param forced: whether the breakpoint should be protected frm usual removal.
        :param module_name: Specify module name to place the BP in (used with `address_or_name` when using a name)
        :param options: can contain an `override` keyword to specify if to override an existing BP
        :return: native LLDB breakpoint
        """
        if address_or_name in [bp.address for bp in self.breakpoints.values()]:
            override = True if options.get('override', True) else False
            if override or prompts.prompt_for_confirmation('A breakpoint already exist in given location. '
                                                           'Would you like to delete the previous one?', True):
                breakpoints = list(self.breakpoints.items())
                for bp_id, bp in breakpoints:
                    if address_or_name == bp.address:
                        self.remove_hilda_breakpoint(bp_id)

        if isinstance(address_or_name, int):
            bp = self.target.BreakpointCreateByAddress(address_or_name)
        elif isinstance(address_or_name, str):
            bp = self.target.BreakpointCreateByName(address_or_name)

        if condition is not None:
            bp.SetCondition(condition)

        # add into Hilda's internal list of breakpoints
        self.breakpoints[bp.id] = HildaClient.Breakpoint(
            address=address_or_name, options=options, forced=forced, callback=callback
        )

        if callback is not None:
            bp.SetScriptCallbackFunction('lldb.hilda_client.bp_callback_router')

        self.log_info(f'Breakpoint #{bp.id} has been set')
        return bp

    def bp_callback_router(self, frame, bp_loc, *_):
        """
        Route the breakpoint callback the specific breakpoint callback.
        :param lldb.SBFrame frame: LLDB Frame object.
        :param lldb.SBBreakpointLocation bp_loc: LLDB Breakpoint location object.
        """
        bp_id = bp_loc.GetBreakpoint().GetID()
        self._bp_frame = frame
        try:
            self.breakpoints[bp_id].callback(self, frame, bp_loc, self.breakpoints[bp_id].options)
        finally:
            self._bp_frame = None

    def show_hilda_breakpoints(self):
        """ Show existing breakpoints created by Hilda. """
        for bp_id, bp in self.breakpoints.items():
            print(f'🚨 Breakpoint #{bp_id}: Forced: {bp.forced}')
            if isinstance(bp.address, int):
                print(f'\tAddress: 0x{bp.address:x}')
            elif isinstance(bp.address, str):
                print(f'\tName: {bp.address}')
            print(f'\tOptions: {bp.options}')

    def save(self, filename=None):
        """
        Save loaded symbols map (for loading later using the load() command)
        :param filename: optional filename for where to store
        """
        if filename is None:
            filename = self._get_saved_state_filename()

        self.log_info(f'saving current state info: {filename}')
        with open(filename, 'wb') as f:
            symbols_copy = {}
            for k, v in self.symbols.items():
                # converting the symbols into serializable objects
                symbols_copy[k] = SerializableSymbol(address=int(v),
                                                     type_=v.type_,
                                                     filename=v.filename)
            pickle.dump(symbols_copy, f)

    def load(self, filename=None):
        """
        Load an existing symbols map (previously saved by the save() command)
        :param filename: filename to load from
        """
        if filename is None:
            filename = self._get_saved_state_filename()

        self.log_info(f'loading current state from: {filename}')
        with open(filename, 'rb') as f:
            symbols_copy = pickle.load(f)

            for k, v in tqdm(symbols_copy.items()):
                self.symbols[k] = self.symbol(v.address)

            # perform sanity test for symbol rand
            if self.symbols.rand() == 0 and self.symbols.rand() == 0:
                # rand returning 0 twice means the loaded file is probably outdated
                raise BrokenLocalSymbolsJarError()

            # assuming the first main image will always change
            self.rebind_symbols(image_range=[0, 0])
            self.init_dynamic_environment()
            self._symbols_loaded = True

    def po(self, expression, cast=None):
        """
        Print given object using LLDB's po command

        Can also run big chunks of native code:

        po('NSMutableString *s = [NSMutableString string]; [s appendString:@"abc"]; [s description]')

        :param expression: either a symbol or string the execute
        :param cast: object type
        :raise EvaluatingExpressionError: LLDB failed to evaluate the expression
        :return: LLDB's po output
        """
        casted_expression = ''
        if cast is not None:
            casted_expression += '(%s)' % cast
        casted_expression += f'0x{expression:x}' if isinstance(expression, int) else str(expression)

        res = lldb.SBCommandReturnObject()
        self.debugger.GetCommandInterpreter().HandleCommand(f'expression -i 0 -lobjc -O -- {casted_expression}', res)
        if not res.Succeeded():
            raise EvaluatingExpressionError(res.GetError())
        return res.GetOutput().strip()

    def globalize_symbols(self):
        """
        Make all symbols in python's global scope
        """
        reserved_names = list(globals().keys()) + dir(builtins)
        for name, value in tqdm(self.symbols.items()):
            if ':' not in name \
                    and '[' not in name \
                    and '<' not in name \
                    and '(' not in name \
                    and '.' not in name:
                self._add_global(name, value, reserved_names)

    def jump(self, symbol: int):
        """ jump to given symbol """
        self.lldb_handle_command(f'j *{symbol}')

    def lldb_handle_command(self, cmd):
        """
        Execute an LLDB command

        For example:
            lldb_handle_command('register read')

        :param cmd:
        """
        self.debugger.HandleCommand(cmd)

    def objc_get_class(self, name: str, module_name: Optional[str] = None) -> objective_c_class.Class:
        """
        Get ObjC class object
        :param module_name:
        :param name:
        :return:
        """
        if module_name is not None:
            ret = self.symbol(self._get_module_class_list(module_name)[name]).objc_class
        else:
            ret = objective_c_class.Class.from_class_name(self, name)
        return ret

    def CFSTR(self, symbol: int) -> Symbol:
        """ Create CFStringRef object from given string """
        return self.cf(symbol)

    def cf(self, data: CfSerializable) -> Symbol:
        """
        Create NSObject from given data (same as ns())
        :param data: Data representing the NSObject, must by JSON serializable
        :return: Pointer to a NSObject
        """
        return self.ns(data)

    def ns(self, data: CfSerializable) -> Symbol:
        """
        Create NSObject from given data (same as cf())
        :param data: Data representing the NSObject, must by JSON serializable
        :return: Pointer to a NSObject
        """
        try:
            json_data = json.dumps({'root': data}, default=self._to_ns_json_default)
        except TypeError as e:
            raise ConvertingToNsObjectError from e

        obj_c_code = (Path(__file__).parent / 'objective_c' / 'to_ns_from_json.m').read_text()
        expression = obj_c_code.replace('__json_object_dump__', json_data.replace('"', r'\"'))
        try:
            return self.evaluate_expression(expression)
        except EvaluatingExpressionError as e:
            raise ConvertingToNsObjectError from e

    def decode_cf(self, address: Union[int, str]) -> CfSerializable:
        """
        Create python object from NS object.
        :param address: NS object.
        :return: Python object.
        """
        obj_c_code = (Path(__file__).parent / 'objective_c' / 'from_ns_to_json.m').read_text()
        address = f'0x{address:x}' if isinstance(address, int) else address
        expression = obj_c_code.replace('__ns_object_address__', address)
        try:
            json_dump = self.po(expression)
        except EvaluatingExpressionError as e:
            raise ConvertingFromNSObjectError from e
        return json.loads(json_dump, object_hook=self._from_ns_json_object_hook)['root']

    def evaluate_expression(self, expression) -> Symbol:
        """
        Wrapper for LLDB's EvaluateExpression.
        Used for quick code snippets.

        Feel free to use local variables inside the expression using format string.
        For example:
            currentDevice = objc_get_class('UIDevice').currentDevice
            evaluate_expression(f'[[{currentDevice} systemName] hasPrefix:@"2"]')

        :param expression:
        :return: returned symbol
        """
        # prepending a prefix so LLDB knows to return an int type
        if isinstance(expression, int):
            formatted_expression = f'(intptr_t)0x{expression:x}'
        else:
            formatted_expression = str(expression)

        options = lldb.SBExpressionOptions()
        options.SetIgnoreBreakpoints(self.configs.evaluation_ignore_breakpoints)
        options.SetTryAllThreads(True)
        options.SetUnwindOnError(self.configs.evaluation_unwind_on_error)

        e = self.frame.EvaluateExpression(formatted_expression, options)

        if not e.error.Success():
            raise EvaluatingExpressionError(str(e.error))

        return self.symbol(e.unsigned)

    def import_module(self, filename: str, name: Optional[str] = None) -> Any:
        """
        Import & reload given python module (intended mainly for external snippets)
        :param filename: Python filename to import
        :param name: Optional module name, or otherwise use the filename
        :return: Python module
        """
        filename = os.path.expanduser(filename)
        if name is None:
            name = os.path.splitext(os.path.basename(filename))[0]
        spec = importlib.util.spec_from_file_location(name, filename)
        m = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(m)
        return m

    def unwind(self) -> bool:
        """ Unwind the stack (useful when get_evaluation_unwind() == False) """
        return self.thread.UnwindInnermostExpression().Success()

    @cached_property
    def pid(self) -> int:
        return self.process.GetProcessID()

    @property
    def thread(self):
        """ Current active thread. """
        if self._bp_frame is not None:
            return self._bp_frame.GetThread()
        return self.process.GetSelectedThread()

    @property
    def frame(self):
        """ Current active frame. """
        if self._bp_frame is not None:
            return self._bp_frame
        return self.thread.GetSelectedFrame()

    @contextmanager
    def stopped(self, interval=0):
        """
        Context-Manager for execution while process is stopped.
        If interval is supplied, then if the device is in running state, it will sleep for the interval
        given before and after execution.
        """
        is_running = self.process.GetState() == lldb.eStateRunning

        if is_running:
            self.stop()
            time.sleep(interval)

        try:
            yield
        finally:
            if is_running:
                time.sleep(interval)
                self.cont()

    @contextmanager
    def safe_malloc(self, size):
        """
        Context-Manager for allocating a block of memory which is freed afterwards
        :param size:
        :return:
        """
        block = self.symbols.malloc(size)
        if block == 0:
            raise IOError(f'failed to allocate memory of size: {size} bytes')

        try:
            yield block
        finally:
            self.symbols.free(block)

    @contextmanager
    def sync_mode(self):
        """ Context-Manager for execution while LLDB is in sync mode. """
        is_async = self.debugger.GetAsync()
        self.debugger.SetAsync(False)
        try:
            yield
        finally:
            self.debugger.SetAsync(is_async)

    def init_dynamic_environment(self):
        """ Init session-scoped process dynamic dependencies. """
        self.log_debug('init dynamic environment')
        self._dynamic_env_loaded = True

        self.log_debug('disable mach_msg receive errors')
        try:
            CFRunLoopServiceMachPort_hooks.disable_mach_msg_errors()
        except SymbolAbsentError:
            self.log_warning('failed to disable mach_msg errors')

        objc_code = """
        @import ObjectiveC;
        @import Foundation;
        """
        try:
            self.po(objc_code)
        except EvaluatingExpressionError:
            # first time is expected to fail. bug in LLDB?
            pass

    def log_warning(self, message):
        """ Log at warning level """
        self.logger.warning(message)

    def log_debug(self, message):
        """ Log at debug level """
        self.logger.debug(message)

    def log_error(self, message):
        """ Log at error level """
        self.logger.error(message)

    def log_critical(self, message):
        """ Log at critical level """
        self.logger.critical(message)
        raise HildaException(message)

    def log_info(self, message):
        """ Log at info level """
        self.logger.info(message)

    def add_lldb_symbol(self, symbol: lldb.SBSymbol) -> Symbol:
        """
        Convert an LLDB symbol into Hilda's symbol object and insert into `symbols` global
        :param symbol: LLDB symbol
        :return: converted symbol
        :raise AddingLldbSymbolError: Hilda failed to convert the LLDB symbol.
        """
        load_addr = symbol.addr.GetLoadAddress(self.target)
        if load_addr == 0xffffffffffffffff:
            # skip those not having a real address
            raise AddingLldbSymbolError()

        name = symbol.name
        type_ = symbol.GetType()

        if name in ('<redacted>',) or (type_ not in (lldb.eSymbolTypeCode,
                                                     lldb.eSymbolTypeRuntime,
                                                     lldb.eSymbolTypeData,
                                                     lldb.eSymbolTypeObjCMetaClass)):
            # ignore unnamed symbols and those which are not in a really used type
            raise AddingLldbSymbolError()

        value = self.symbol(load_addr)

        # add it into symbols global
        self.symbols[name] = value
        self.symbols[f'{name}{{{value.filename}}}'] = value

        return value

    def interact(self, additional_namespace: Optional[typing.Mapping] = None,
                 startup_files: Optional[List[str]] = None) -> None:
        """ Start an interactive Hilda shell """
        if not self._dynamic_env_loaded:
            self.init_dynamic_environment()
        print('\n')
        self.log_info(html_to_ansi(GREETING))
        ipython_config = Config()
        ipython_config.IPCompleter.use_jedi = True
        ipython_config.BaseIPythonApplication.profile = 'hilda'
        ipython_config.InteractiveShellApp.extensions = ['hilda.ipython_extensions.magics',
                                                         'hilda.ipython_extensions.events',
                                                         'hilda.ipython_extensions.keybindings']
        ipython_config.InteractiveShellApp.exec_lines = ["disable_logs()"]
        if startup_files is not None:
            ipython_config.InteractiveShellApp.exec_files = startup_files
            self.log_debug(f'Startup files - {startup_files}')

        namespace = globals()
        namespace.update(locals())
        namespace['p'] = self
        if additional_namespace is not None:
            namespace.update(additional_namespace)
        sys.argv = ['a']
        IPython.start_ipython(config=ipython_config, user_ns=namespace)
        self.detach()

    @staticmethod
    def _add_global(name: str, value: Any, reserved_names=None):
        if reserved_names is None or name not in reserved_names:
            # don't override existing symbols
            globals()[name] = value

    @staticmethod
    def _get_saved_state_filename():
        return '/tmp/cache.hilda'

    @staticmethod
    def _to_ns_json_default(obj):
        if isinstance(obj, bytes):
            return f'__hilda_magic_key__|NSData|{base64.b64encode(obj).decode()}'
        elif isinstance(obj, datetime):
            return f'__hilda_magic_key__|NSDate|{obj.timestamp()}'
        raise TypeError

    @staticmethod
    def _from_ns_json_object_hook(obj: dict):
        parsed_object = {}
        for key, value in obj.items():
            parsed_object[HildaClient._from_ns_parse_function(key)] = HildaClient._from_ns_parse_function(value)
        return parsed_object

    @staticmethod
    def _from_ns_parse_function(obj):
        if isinstance(obj, list):
            return list(map(HildaClient._from_ns_parse_function, obj))
        if not isinstance(obj, str) or not obj.startswith('__hilda_magic_key__'):
            return obj
        _, type_, data = obj.split('|')
        if type_ == 'NSData':
            return base64.b64decode(data)
        if type_ == 'NSDictionary':
            return tuple(json.loads(data, object_hook=HildaClient._from_ns_json_object_hook).items())
        if type_ == 'NSArray':
            return tuple(json.loads(data, object_hook=HildaClient._from_ns_json_object_hook))
        if type_ == 'NSNumber':
            return eval(data)
        if type_ == 'NSNull':
            return None
        if type_ == 'NSDate':
            return datetime.fromtimestamp(eval(data), timezone.utc)

    def _serialize_call_params(self, argv):
        args_conv = []
        for arg in argv:
            if isinstance(arg, str) or isinstance(arg, bytes):
                if isinstance(arg, str):
                    arg = arg.encode()
                arg = ''.join([f'\\x{b:02x}' for b in arg])
                args_conv.append(f'(intptr_t)"{arg}"')
            elif isinstance(arg, int) or isinstance(arg, Symbol):
                arg = int(arg) & 0xffffffffffffffff
                args_conv.append(f'0x{arg:x}')
            else:
                raise NotImplementedError('cannot serialize argument')
        return args_conv

    def _generate_call_expression(self, address, params):
        args_type = ','.join(['intptr_t'] * len(params))
        args_conv = ','.join(params)

        if self.arch == 'arm64e':
            address = f'ptrauth_sign_unauthenticated((void *){address}, ptrauth_key_asia, 0)'

        return f'((intptr_t(*)({args_type}))({address}))({args_conv})'

    @staticmethod
    def _std_string(value):
        if struct.unpack("b", (value + 23).peek(1))[0] >= 0:
            return value.peek_str()
        else:
            return value[0].peek_str()

    def _monitor_format_value(self, fmt, value):
        if callable(fmt):
            return fmt(self, value)
        formatters = {
            'x': lambda val: f'0x{int(val):x}',
            's': lambda val: val.peek_str() if val else None,
            'cf': lambda val: val.cf_description,
            'po': lambda val: val.po(),
            'std::string': self._std_string
        }
        if fmt in formatters:
            return formatters[fmt](value)
        else:
            return f'{value:x} (unsupported format)'

    @cached_property
    def _ks(self) -> Optional['Ks']:
        if not lldb.KEYSTONE_SUPPORT:
            return False
        platforms = {'arm64': Ks(KS_ARCH_ARM64, KS_MODE_LITTLE_ENDIAN),
                     'arm64e': Ks(KS_ARCH_ARM64, KS_MODE_LITTLE_ENDIAN),
                     'x86_64h': Ks(KS_ARCH_X86, KS_MODE_64)}
        return platforms.get(self.arch)

    def _get_module_class_list(self, module_name: str):
        for m in self.target.module_iter():
            if module_name != m.file.basename:
                continue
            objc_classlist = m.FindSection('__DATA').FindSubSection('__objc_classlist')
            objc_classlist_addr = self.symbol(objc_classlist.GetLoadAddress(self.target))
            obj_c_code = (Path(__file__).parent / 'objective_c' / 'get_objectivec_class_by_module.m').read_text()
            obj_c_code = obj_c_code.replace('__count_objc_class', f'{objc_classlist.size // 8}').replace(
                '__objc_class_list',
                f'{objc_classlist_addr}')
            return json.loads(self.po(obj_c_code))



from construct import Computed, Hex, Int32ul, LazyArray, Struct, Tell, this

from hilda.snippets.macho.macho_load_commands import LOAD_COMMAND_TYPE, load_command_t


def __calculate_aslr(ctx):
    loaded_address = ctx.load_address
    for load_command in ctx.load_commands:
        if load_command.cmd == LOAD_COMMAND_TYPE.LC_SEGMENT_64 and load_command.data.segname == '__TEXT':
            return loaded_address - load_command.data.vmaddr


mach_header_t = Struct(
    'load_address' / Tell,
    'magic' / Hex(Int32ul),
    'cputype' / Hex(Int32ul),
    'cpusubtype' / Hex(Int32ul),
    'filetype' / Hex(Int32ul),
    'ncmds' / Hex(Int32ul),
    'sizeofcmds' / Hex(Int32ul),
    'flags' / Hex(Int32ul),
    'reserved' / Hex(Int32ul),
    'load_commands' / LazyArray(this.ncmds, load_command_t),
    'aslr' / Computed(__calculate_aslr)
)



from pprint import pformat

from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import PythonLexer

from hilda.exceptions import ConvertingFromNSObjectError
from hilda.lldb_importer import lldb

# module global for storing all active xpc connections
active_xpc_connections = {}


def safe_monitor(symbol_name, **kwargs):
    hilda = lldb.hilda_client
    try:
        hilda.symbols[symbol_name].monitor(**kwargs)
    except KeyError:
        args = []
        for k, v in kwargs.items():
            args.append(f'{k}={v}')
        lldb.hilda_client.log_warning(f'missing symbol: "{symbol_name}". try to locate it manually (in IDA) and place '
                                      f'a breakpoint as follows:\n\tfile_symbol(ADDRESS).monitor({", ".join(args)})')


def sniff_send():
    safe_monitor('xpc_connection_send_message{libxpc.dylib}', regs={'x0': 'po', 'x1': 'po'})
    safe_monitor('xpc_connection_send_message_with_reply{libxpc.dylib}', regs={'x0': 'po', 'x1': 'po'})
    safe_monitor('xpc_connection_send_message_with_reply_sync{libxpc.dylib}', regs={'x0': 'po', 'x1': 'po'})


def sniff_receive():
    safe_monitor('_xpc_connection_call_event_handler', regs={'x0': 'po', 'x1': 'po'})


def sniff_incoming_event():
    safe_monitor('__XPC_CONNECTION_EVENT_HANDLER_CALLOUT__', regs={'x0': 'po', 'x1': 'po'})


def sniff_activities():
    safe_monitor('__XPC_ACTIVITY_CALLING_HANDLER__', regs={'x0': 'po'})


def sniff_all():
    sniff_send()
    sniff_receive()


def from_xpc_object(address: int):
    """
    Convert XPC object to python object.
    :param address: Address of XPC object.
    """
    return lldb.hilda_client.decode_cf(f'_CFXPCCreateCFObjectFromXPCObject({address})')


def disable_transaction_exit() -> None:
    """
    xpc_transaction_exit_clean will kill the process when transaction is done.
    By patching this function the process will stay alive.
    """
    hilda = lldb.hilda_client
    hilda.symbols.xpc_transaction_exit_clean.poke(b'\xc0\x03\x5f\xd6')  # ret


def to_xpc_object(obj: object):
    """
    Convert python object to XPC object.
    :param obj: Native python object
    """
    hilda = lldb.hilda_client
    return hilda.symbols._CFXPCCreateXPCObjectFromCFObject(hilda.ns(obj))


def xpc_to_python_monitor_format(hilda_client, address):
    """
    Format an XPC object as a python object, intended to use as a callback to monitor command.
    It depends on the object compatibility with CF object.
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param hilda.symbol.Symbol address: Symbol to format.
    """
    try:
        formatted = pformat(from_xpc_object(address))
        return highlight(formatted, PythonLexer(), TerminalTrueColorFormatter(style='native'))
    except ConvertingFromNSObjectError:
        return address.po()


def send_message_raw(service_name, message_raw):
    hilda = lldb.hilda_client

    with hilda.stopped():
        if service_name in active_xpc_connections:
            conn = active_xpc_connections[service_name]
        else:
            # 0 is for connecting instead of listening
            conn = hilda.symbols.xpc_connection_create_mach_service(service_name, 0, 0)
            assert conn != 0, 'failed to create xpc connection'
            hilda.po(f'''xpc_connection_set_event_handler({conn}, ^(xpc_object_t obj) {{}})''')
            hilda.symbols.xpc_connection_resume(conn)
            active_xpc_connections[service_name] = conn

        result = hilda.symbols.xpc_connection_send_message_with_reply_sync(conn, message_raw)
        return result


def send_message(service_name, message: object):
    hilda = lldb.hilda_client

    with hilda.stopped():
        message_raw = to_xpc_object(message)
        assert message_raw != 0, 'failed to convert python message object to native xpc object'
        return from_xpc_object(send_message_raw(service_name, message_raw))



import json
from collections import OrderedDict
from pathlib import Path

from hilda.ui.views import BackTraceView, DisassemblyView, RegistersView, StackView


class DotDict(OrderedDict):
    """ Extends `OrderedDict` to access items using dot """
    __slots__ = ["__recursion_lock__"]

    def __getattr__(self, name):
        try:
            if name in self.__slots__:
                return object.__getattribute__(self, name)
            else:
                return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        try:
            if name in self.__slots__:
                return object.__setattr__(self, name, value)
            else:
                self[name] = value
        except KeyError:
            raise AttributeError(name)


class ColorScheme(DotDict):
    def __init__(self, scheme: Path):
        with scheme.open('r') as f:
            data = json.load(f)
        super().__init__(data)


class ViewsManager(DotDict):
    pass


class UiManager:
    """
    Manages Hilda's UI.

    - New views can be added to `ViewManager` (Must be inherent `Views`)
    - Colors can be changed from terminal (Setting `ui_manager.color.title = 'green'` or be modifying `colors.json`)
    - Views can be disabled from terminal (Setting `ui_manager.color.title=False`)
    """

    def __init__(self, hilda_client, scheme_file=Path(__file__).parent / 'colors.json',
                 active: bool = True):
        """
        :param hilda_client: hilda.hilda_client.HildaClient
        :param scheme_file: pathlib.Path
        :param active: bool
        """
        self.colors = ColorScheme(scheme_file)
        self.views = ViewsManager({
            'registers': RegistersView(hilda_client, self.colors),
            'disassembly': DisassemblyView(hilda_client, self.colors),
            'stack': StackView(hilda_client, self.colors),
            'backtrace': BackTraceView(hilda_client, self.colors)
        })
        self.active = active

    def show(self, *args) -> None:
        fmt_parts = []

        for view in self.views.values():
            if view.active is False:
                continue
            fmt_parts.append(str(view))
        print('\n'.join(fmt_parts))



from datetime import datetime, timezone

import pytest

from hilda.exceptions import ConvertingToNsObjectError


@pytest.mark.parametrize('source', [
    '',
    '3123123',
    'asdsdasd',
    '12312sadasd',
    'The quick brown fox jumps over the lazy frog123',
])
def test_cfstr(hilda_client, source: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Python string to be converted to CFSTR.
    """
    cfstr = hilda_client.CFSTR(source)
    assert cfstr
    po = cfstr.po()
    assert po == source or (not source and po == '<object returned empty description>')
    assert source in cfstr.cf_description
    assert 'NSString' in list(map(lambda sup: sup.name, cfstr.objc_class.iter_supers()))


def test_ns_data(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    data = b'\x01\x00asdasd\xff\xfe58'
    ns_data = hilda_client.ns(data).objc_symbol
    assert ns_data.bytes.read(len(data)) == data


@pytest.mark.parametrize('day, month, year', [(1, 1, 1970), (11, 10, 2021)])
def test_ns_date(hilda_client, day: int, month: int, year: int):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param day: Date day.
    :param month: Date month.
    :param year: Date year.
    """
    date_po = f'{year:04d}-{month:02d}-{day:02d} 00:00:00 +0000'
    assert hilda_client.ns(datetime(day=day, month=month, year=year, tzinfo=timezone.utc)).po() == date_po


@pytest.mark.parametrize('source, result', [
    ({'asdasd': 234234, 234234: 'asdasd', 1: True, 'a': [False, False]}, '{1=1;234234=asdasd;a=(0,0);asdasd=234234;}'),
    ({}, '{}'),
    ({'asdasds': 324234}, '{asdasds=324234;}'),
    ({1: {2: {3: 'a'}}}, '{1={2={3=a;};};}'),
    ({1: b'\x00\x01'}, '{1={length=2,bytes=0x0001};}'),
    ({1: datetime(1970, 1, 1, tzinfo=timezone.utc)}, '{1="1970-01-0100:00:00+0000";}'),
])
def test_ns_dict(hilda_client, source: dict, result: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Python object to be converted to NS object.
    :param result: PO result.
    """
    ns_object = hilda_client.ns(source)
    assert ''.join(ns_object.po().split()) == result
    assert 'NSDictionary' in list(map(lambda sup: sup.name, ns_object.objc_class.iter_supers()))


def test_ns_none(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    ns_object = hilda_client.ns(None)
    assert ns_object.po() == '<null>'
    assert ns_object.objc_class.name == 'NSNull'


@pytest.mark.parametrize('source, result', [
    (['asdasd', 234234, 1, True, {'a': False}], '(asdasd,234234,1,1,{a=0;})'),
    ([], '()'),
    (['asdasds', 324234], '(asdasds,324234)'),
    ([1, [2, [3, 'a']]], '(1,(2,(3,a)))'),
    (['asdasds', b'324234'], '(asdasds,{length=6,bytes=0x333234323334})'),
    (['asdasds', datetime(2014, 4, 1, tzinfo=timezone.utc)], '(asdasds,"2014-04-0100:00:00+0000")'),
])
def test_ns_array(hilda_client, source, result: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Python object to be converted to NS object.
    :param result: PO result.
    """
    ns_object = hilda_client.ns(source)
    assert ''.join(ns_object.cf_description.split()) == result
    assert 'NSArray' in list(map(lambda sup: sup.name, ns_object.objc_class.iter_supers()))


@pytest.mark.parametrize('source', [0, 1, -1, 1.5, -1.5, 0xfffffffffffffffffff, -0xfffffffffffffffffff, 1 / 3])
def test_ns_number(hilda_client, source):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Number to convert to NS object.
    """
    ns_object = hilda_client.ns(source)
    assert ns_object.po() == str(source)
    assert 'NSNumber' in list(map(lambda sup: sup.name, ns_object.objc_class.iter_supers()))


@pytest.mark.parametrize('source', [
    object(),
    {'aa', 123},
])
def test_error_converting_to_ns(hilda_client, source):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Python object to convert to NS object.
    """
    with pytest.raises(ConvertingToNsObjectError):
        hilda_client.ns(source)



import json
import time
from collections import namedtuple
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import Any
from uuid import uuid4

from objc_types_decoder.decode import decode as decode_type
from objc_types_decoder.decode import decode_with_tail
from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import ObjectiveCLexer

from hilda.exceptions import GettingObjectiveCClassError
from hilda.symbols_jar import SymbolsJar

Ivar = namedtuple('Ivar', 'name type_ offset')
Property = namedtuple('Property', 'name attributes')
PropertyAttributes = namedtuple('PropertyAttributes', 'synthesize type_ list')


def convert_encoded_property_attributes(encoded):
    conversions = {
        'R': lambda x: 'readonly',
        'C': lambda x: 'copy',
        '&': lambda x: 'strong',
        'N': lambda x: 'nonatomic',
        'G': lambda x: 'getter=' + x[1:],
        'S': lambda x: 'setter=' + x[1:],
        'd': lambda x: 'dynamic',
        'W': lambda x: 'weak',
        'P': lambda x: '<garbage-collected>',
        't': lambda x: 'encoding=' + x[1:],
    }

    type_, tail = decode_with_tail(encoded[1:])
    attributes = []
    synthesize = None
    for attr in filter(None, tail.lstrip(',').split(',')):
        if attr[0] in conversions:
            attributes.append(conversions[attr[0]](attr))
        elif attr[0] == 'V':
            synthesize = attr[1:]

    return PropertyAttributes(type_=type_, synthesize=synthesize, list=attributes)


@dataclass
class Method:
    name: str
    client: Any = field(compare=False)
    address: int = field(compare=False)
    imp: int = field(compare=False)
    type_: str = field(compare=False)
    return_type: str = field(compare=False)
    is_class: bool = field(compare=False)
    args_types: list = field(compare=False)

    @staticmethod
    def from_data(data: dict, client):
        """
        Create Method object from raw data.
        :param data: Data as loaded from get_objectivec_symbol_data.m.
        :param hilda.hilda_client.HildaClient client: Hilda client.
        """
        return Method(
            name=data['name'],
            client=client,
            address=client.symbol(data['address']),
            imp=client.symbol(data['imp']),
            type_=data['type'],
            return_type=decode_type(data['return_type']),
            is_class=data['is_class'],
            args_types=list(map(decode_type, data['args_types']))
        )

    def set_implementation(self, new_imp: int):
        self.client.symbols.method_setImplementation(self.address, new_imp)
        self.imp = self.client.symbol(new_imp)

    def __str__(self):
        if ':' in self.name:
            args_names = self.name.split(':')
            name = ' '.join(['{}:({})'.format(*arg) for arg in zip(args_names, self.args_types[2:])])
        else:
            name = self.name
        prefix = '+' if self.is_class else '-'
        return f'{prefix} {name}; // 0x{self.address:x} (returns: {self.return_type})\n'


class Class(object):
    """
    Wrapper for ObjectiveC Class object.
    """

    def __init__(self, client, class_object=0, class_data: dict = None):
        """
        :param hilda.hilda_client.HildaClient client:
        :param hilda.symbol.Symbol class_object:
        """
        self._client = client
        self._class_object = class_object
        self.protocols = []
        self.ivars = []
        self.properties = []
        self.methods = []
        self.name = ''
        self.super = None
        if class_data is None:
            self.reload()
        else:
            self._load_class_data(class_data)

    @staticmethod
    def from_class_name(client, class_name: str):
        """
        Create ObjectiveC Class from given class name.
        :param hilda.hilda_client.HildaClient client: Hilda client.
        :param class_name: Class name.
        """
        obj_c_code = (Path(__file__).parent / 'objective_c' / 'get_objectivec_class_description.m').read_text()
        obj_c_code = obj_c_code.replace('__class_address__', '0').replace('__class_name__', class_name)
        class_symbol = Class(client, class_data=json.loads(client.po(obj_c_code)))
        if class_symbol.name != class_name:
            raise GettingObjectiveCClassError()
        return class_symbol

    @staticmethod
    def sanitize_name(name: str):
        """
        Sanitize python name to ObjectiveC name.
        """
        if name.startswith('_'):
            name = '_' + name[1:].replace('_', ':')
        else:
            name = name.replace('_', ':')
        return name

    def reload(self):
        """
        Reload class object data.
        Should be used whenever the class layout changes (for example, during method swizzling)
        """
        obj_c_code = (Path(__file__).parent / 'objective_c' / 'get_objectivec_class_description.m').read_text()
        obj_c_code = obj_c_code.replace('__class_address__', f'{self._class_object:d}')
        obj_c_code = obj_c_code.replace('__class_name__', self.name)
        self._load_class_data(json.loads(self._client.po(obj_c_code)))

    def show(self):
        """
        Print to terminal the highlighted class description.
        """
        print(highlight(str(self), ObjectiveCLexer(), TerminalTrueColorFormatter(style='native')))

    def objc_call(self, sel: str, *args):
        """
        Invoke a selector on the given class object.
        :param sel: Selector name.
        :return: whatever the selector returned as a symbol.
        """
        return self._class_object.objc_call(sel, *args)

    def get_method(self, name: str):
        """
        Get a specific method implementation.
        :param name: Method name.
        :return: Method.
        """
        for method in self.methods:
            if method.name == name:
                return method

    def capture_self(self, sync: bool = False):
        """
        Capture the first called `self` from this class.
        Access using `self.captured_objects`
        :param sync: Should wait until captured object is returned?
        :return: Captured object if sync is True, None otherwise
        """
        class_name = self.name

        if class_name in self._client.captured_objects:
            del self._client.captured_objects[class_name]

        def hook(hilda, frame, bp_loc, options):
            hilda.log_info(f'self object has been captured from {options["name"]}')
            hilda.log_info('removing breakpoints')
            for bp_id, bp in list(hilda.breakpoints.items()):
                if 'group_uuid' in bp.options and bp.options.get('group_uuid', '') == options['group_uuid']:
                    hilda.remove_hilda_breakpoint(bp_id)
            captured = hilda.evaluate_expression('$arg1')
            captured = captured.objc_symbol
            hilda.captured_objects[options['name'].split(' ')[0].split('[')[1]] = captured
            hilda.cont()

        group_uuid = str(uuid4())

        for method in self.methods:
            if not method.is_class:
                # only instance methods are relevant for capturing self
                method.imp.bp(hook, group_uuid=group_uuid,
                              name=f'-[{class_name} {method.name}]')

        if sync:
            self._client.cont()
            self._client.log_debug('Waiting for desired object to be captured...')
            while class_name not in self._client.captured_objects:
                time.sleep(1)

            return self._client.captured_objects[class_name]

    def monitor(self, **kwargs):
        """
        Proxy for monitor command.
        """
        self.symbols_jar.monitor(**kwargs)

    def bp(self, callback=None, **kwargs):
        """
        Proxy for bp command.
        """
        for method in self.methods:
            kwargs['name'] = f'[{self.name} {method.name}]'
            method.imp.bp(callback, **kwargs)

    def iter_supers(self):
        """
        Iterate over the super classes of the class.
        """
        sup = self.super
        while sup is not None:
            yield sup
            sup = sup.super

    def _load_class_data(self, data: dict):
        self._class_object = self._client.symbol(data['address'])
        self.super = Class(self._client, data['super']) if data['super'] else None
        self.name = data['name']
        self.protocols = data['protocols']
        self.ivars = [
            Ivar(name=ivar['name'], type_=decode_type(ivar['type']) if ivar['type'] else 'unknown_type_t',
                 offset=ivar['offset'])
            for ivar in data['ivars']
        ]
        self.properties = [
            Property(name=prop['name'], attributes=convert_encoded_property_attributes(prop['attributes']))
            for prop in data['properties']
        ]
        self.methods = [Method.from_data(method, self._client) for method in data['methods']]

    @property
    def symbols_jar(self) -> SymbolsJar:
        """ Get a SymbolsJar object for quick operations on all methods """
        jar = SymbolsJar.create(self._client)

        for m in self.methods:
            jar[f'[{self.name} {m.name}]'] = m.imp

        return jar

    def __dir__(self):
        result = set()

        for method in self.methods:
            if method.is_class:
                result.add(method.name.replace(':', '_'))

        for sup in self.iter_supers():
            if self._client.configs.nsobject_exclusion and sup.name == 'NSObject':
                continue
            for method in sup.methods:
                if method.is_class:
                    result.add(method.name.replace(':', '_'))

        result.update(list(super(Class, self).__dir__()))
        return list(result)

    def __str__(self):
        protocol_buf = f'<{",".join(self.protocols)}>' if self.protocols else ''

        if self.super is not None:
            buf = f'@interface {self.name}: {self.super.name} {protocol_buf}\n'
        else:
            buf = f'@interface {self.name} {protocol_buf}\n'

        # Add ivars
        buf += '{\n'
        for ivar in self.ivars:
            buf += f'\t{ivar.type_} {ivar.name}; // 0x{ivar.offset:x}\n'
        buf += '}\n'

        # Add properties
        for prop in self.properties:
            buf += f'@property ({",".join(prop.attributes.list)}) {prop.attributes.type_} {prop.name};\n'

            if prop.attributes.synthesize is not None:
                buf += f'@synthesize {prop.name} = {prop.attributes.synthesize};\n'

        # Add methods
        for method in self.methods:
            buf += str(method)

        buf += '@end'
        return buf

    def __repr__(self):
        return f'<objC Class "{self.name}">'

    def __getitem__(self, item):
        for method in self.methods:
            if method.name == item:
                if method.is_class:
                    return partial(self.objc_call, item)
                else:
                    raise AttributeError(f'{self.name} class has an instance method named {item}, '
                                         f'not a class method')

        for sup in self.iter_supers():
            for method in sup.methods:
                if method.name == item:
                    if method.is_class:
                        return partial(self.objc_call, item)
                    else:
                        raise AttributeError(f'{self.name} class has an instance method named {item}, '
                                             f'not a class method')

        raise AttributeError(f''''{self.name}' class has no attribute {item}''')

    def __getattr__(self, item: str):
        return self[self.sanitize_name(item)]



from datetime import datetime, timezone

import pytest

from hilda.exceptions import ConvertingFromNSObjectError


@pytest.mark.parametrize('source', [
    '',
    '3123123',
    'asdsdasd',
    '12312sadasd',
    'The quick brown fox jumps over the lazy frog123',
])
def test_cfstr(hilda_client, source: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: CFSTR to be converted to Python string.
    """
    cfstr = hilda_client.evaluate_expression(f'@"{source}"')
    assert cfstr
    assert hilda_client.decode_cf(cfstr) == source


def test_ns_data(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    data = b'\x01\x00asdasd\xff\xfe58'
    with hilda_client.safe_malloc(len(data)) as buffer:
        buffer.write(data)
        ns_data = hilda_client.evaluate_expression(f'[NSData dataWithBytes:(char *)0x{buffer:x} length:{len(data)}]')
        assert hilda_client.decode_cf(ns_data) == data


def test_ns_data_in_dict(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    data = b'\x01\x00asdasd\xff\xfe58'
    with hilda_client.safe_malloc(len(data)) as buffer:
        buffer.write(data)
        ns_dict = hilda_client.evaluate_expression(
            f'@{{"a": [NSData dataWithBytes:(char *)0x{buffer:x} length:{len(data)}]}}'
        )
        assert hilda_client.decode_cf(ns_dict) == {"a": data}


def test_ns_data_in_array(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    data = b'\x01\x00asdasd\xff\xfe58'
    with hilda_client.safe_malloc(len(data)) as buffer:
        buffer.write(data)
        ns_dict = hilda_client.evaluate_expression(
            f'@[[NSData dataWithBytes:(char *)0x{buffer:x} length:{len(data)}]]'
        )
        assert hilda_client.decode_cf(ns_dict) == [data]


@pytest.mark.parametrize('day, month, year', [(1, 1, 1970), (11, 10, 2021)])
def test_ns_date(hilda_client, day: int, month: int, year: int):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param day: Date day.
    :param month: Date month.
    :param year: Date year.
    """
    date = hilda_client.evaluate_expression(f'''
        NSDateComponents *comps = [NSDateComponents new];
        [comps setDay:{day}];
        [comps setMonth:{month}];
        [comps setYear:{year}];
        [comps setTimeZone:[NSTimeZone timeZoneWithAbbreviation:@"UTC"]];
        [[NSCalendar currentCalendar] dateFromComponents:comps];
    ''')
    assert hilda_client.decode_cf(date) == datetime(day=day, month=month, year=year, tzinfo=timezone.utc)


@pytest.mark.parametrize('source, result', [
    # Dictionaries
    ('@{@1:@1,@234234:@"asdasd",@"a":@[@0,@0],@"asdasd":@234234}',
     {'asdasd': 234234, 234234: 'asdasd', 1: 1, 'a': [0, 0]}),
    ('@{}', {}),
    ('@{"asdasds":324234}', {'asdasds': 324234}),
    ('@{[NSNull null]:324234}', {None: 324234}),
    ('@{@{"a":1}:324234}', {(("a", 1),): 324234}),
    ('@{@["a",1]:324234}', {("a", 1): 324234}),
    ('@{1:@{2:@{3:"a"}}}', {1: {2: {3: 'a'}}}),
    # Arrays
    ('@[@"asdasd",@234234,@1,@1,@{@"a":@0}]', ['asdasd', 234234, 1, 1, {'a': 0}]),
    ('@[]', []),
    ('@["asdasds",324234]', ['asdasds', 324234]),
    ('@[1,@[2,@[3,"a"]]]', [1, [2, [3, 'a']]]),
    # Sets
    ('[NSSet setWithArray: @[@"1", @42]]', ['1', 42])
])
def test_ns_nested_objects(hilda_client, source: str, result):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: NS object expression to be converted to Python object.
    :param result: Python object.
    """
    ns_object = hilda_client.evaluate_expression(source)
    assert ns_object
    assert hilda_client.decode_cf(ns_object) == result


def test_ns_none(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    ns_object = hilda_client.evaluate_expression('[NSNull null]')
    assert hilda_client.decode_cf(ns_object) is None


@pytest.mark.parametrize('source', [0, 1, -1, 1.5, -1.5, 0xfffffffffffffffffff, -0xfffffffffffffffffff, 1 / 3])
def test_ns_number(hilda_client, source):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Number to convert from NS object.
    """
    ns_number = hilda_client.evaluate_expression(f'[NSDecimalNumber decimalNumberWithString:@"{source}"]')
    assert hilda_client.decode_cf(ns_number) == source


@pytest.mark.parametrize('source', [
    '[NSObject new]',
    '0x33',
])
def test_error_converting_from_ns(hilda_client, source):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Expression to convert to python object.
    """
    with pytest.raises(ConvertingFromNSObjectError):
        hilda_client.decode_cf(hilda_client.evaluate_expression(source))



from construct import CString, If, Int64ul, Pointer, Struct, this

from hilda.lldb_importer import lldb
from hilda.snippets.macho.macho import mach_header_t
from hilda.snippets.macho.macho_load_commands import LoadCommands
from hilda.snippets.uuid import uuid_t
from hilda.symbol import SymbolFormatField

dyld_image_info_t = Struct(
    '_imageLoadAddress' / SymbolFormatField(lldb.hilda_client),
    'imageLoadAddress' / Pointer(this._imageLoadAddress, mach_header_t),
    '_imageFilePath' / SymbolFormatField(lldb.hilda_client),
    'imageFilePath' / Pointer(this._imageFilePath, CString('utf8')),
    'imageFileModDate' / Int64ul
)

dyld_uuid_info_t = Struct(
    '_imageLoadAddress' / SymbolFormatField(lldb.hilda_client),
    'imageLoadAddress' / If(this._imageLoadAddress != 0, Pointer(this._imageLoadAddress,
                                                                 mach_header_t)),
    'imageUUID' / uuid_t
)


class ImageInfo(object):
    def __init__(self, image_info_data):
        self.__image_info_data = image_info_data
        self.__file_path = image_info_data.imageFilePath
        self.__image_mach_header = image_info_data.imageLoadAddress
        self.__load_commands = None

    @property
    def file_path(self):
        return self.__file_path

    @property
    def load_commands(self):
        if self.__load_commands is None:
            self.__load_commands = LoadCommands(self.__image_mach_header.load_commands)

        return self.__load_commands

    def __str__(self):
        return f"<ImageInfo: {self.file_path}>"

    def __repr__(self):
        return self.__str__()



from hilda.lldb_importer import lldb

TLV_MAP = {
    0x00: 'METHOD',
    0x01: 'IDENTIFIER',
    0x02: 'SALT',
    0x03: 'PUBLIC_KEY',
    0x04: 'PROOF',
    0x05: 'ENCRYPTED_DATA',
    0x06: 'STATE',
    0x07: 'ERROR',
    0x08: 'RETRY_DELAY',
    0x09: 'CERTIFICATE',
    0x0a: 'SIGNATURE',
    0x0b: 'PERMISSIONS',
    0x0c: 'FRAGMENT_DATA',
    0x0d: 'FRAGMENT_LAST',
    0x0e: 'SESSION_ID',
    0x0f: 'TTL',
    0x10: 'EXTRA_DATA',
    0x11: 'INFO',
    0x12: 'ACL',
    0x13: 'FLAGS',
    0x14: 'VALIDATION_DATA',
    0x15: 'MFI_AUTH_TOKEN',
    0x16: 'MFI_PRODUCT_TYPE',
    0x17: 'SERIAL_NUMBER',
    0x18: 'MFI_AUTH_TOKEN_UUID',
    0x19: 'APP_FLAGS',
    0x1a: 'OWNERSHIP_PROOF',
    0x1b: 'SETUP_CODE_TYPE',
    0x1c: 'PRODUCTION_DATA',
    0x1d: 'APP_INFO',
    0xff: 'SEPARATOR'
}


def _TLV8CopyCoalesced_bp(hilda, *args):
    type_ = hilda.registers.x2
    out_len = hilda.registers.x3
    out_len.item_size = 4
    hilda.finish()

    if hilda.registers.x0 == 0:
        print(f'TLV8CopyCoalesced\n'
              f'    type_: {TLV_MAP[type_]}\n'
              f'    buffer: Null\n'
              f'---\n')
    else:
        print(f'TLV8CopyCoalesced\n'
              f'    type_: {TLV_MAP[type_]}\n'
              f'    buffer: {hilda.registers.x0.peek(out_len[0])}\n'
              f'---\n')
    hilda.cont()


def _TLV8BufferAppend_bp(hilda, *args):
    buffer = hilda.registers.x0
    type_ = hilda.registers.x1
    buffer = hilda.registers.x2
    buffer_len = hilda.registers.x3
    print(f'TLV8BufferAppend\n'
          f'    type_: {TLV_MAP[type_]}\n'
          f'    buffer: {buffer.peek(buffer_len)}\n'
          f'---\n')
    hilda.cont()


def _SRPClientStart_libsrp_bp(hilda, *args):
    username = hilda.registers.x2
    username_len = hilda.registers.x3
    password = hilda.registers.x4
    password_len = hilda.registers.x5
    salt = hilda.registers.x6
    salt_len = hilda.registers.x7
    server_public_key = hilda.registers.sp[0]
    server_public_key_len = hilda.registers.sp[1]
    client_public_key = hilda.registers.sp[2]
    client_public_key_len = hilda.registers.sp[3]
    print(f'SRPClientStart_libsrp\n'
          f'    username: {username.peek(username_len)}\n'
          f'    password: {password.peek(password_len)}\n'
          f'    salt: {salt.peek(salt_len)}\n'
          f'    server_public_key: {server_public_key.peek(server_public_key_len)}\n'
          f'    client_public_key: {client_public_key.peek(client_public_key_len)}\n'
          f'---\n')
    hilda.cont()


def _cced25519_sign_bp(hilda, *args):
    sig = hilda.registers.x1
    len = hilda.registers.x2
    msg = hilda.registers.x3
    msg = msg.peek(len)
    public_key = hilda.registers.x4.peek(32)
    private_key = hilda.registers.x5.peek(32)
    hilda.finish()
    print(f'cced25519_sign\n'
          f'    public_key: {public_key}\n'
          f'    private_key: {private_key}\n'
          f'    msg: {msg}\n'
          f'    sig: {sig.peek(64)}\n'
          f'---\n')
    hilda.cont()


def _cced25519_verify_bp(hilda, *args):
    sig = hilda.registers.x3
    len = hilda.registers.x1
    msg = hilda.registers.x2
    msg = msg.peek(len)
    public_key = hilda.registers.x4.peek(32)
    hilda.finish()
    print(f'cced25519_verify\n'
          f'    public_key: {public_key}\n'
          f'    msg: {msg}\n'
          f'    sig: {sig.peek(64)}\n'
          f'---\n')
    hilda.cont()


def _CryptoHKDF_bp(hilda, *args):
    descriptor = hilda.registers.x0
    key = hilda.registers.x1.peek(hilda.registers.x2)
    salt = hilda.registers.x3
    salt_len = hilda.registers.x4
    info = hilda.registers.x5
    info_len = hilda.registers.x6
    out_key = hilda.registers.sp[0]
    hilda.finish()
    print(f'CryptoHKDF\n'
          f'    descriptor: {descriptor}\n'
          f'    key: {key}\n'
          f'    salt: {salt.peek(salt_len)}\n'
          f'    info: {info.peek(info_len)}\n'
          f'    outKey: {out_key.peek(32)}\n'
          f'---\n')
    hilda.cont()


def _chacha20_poly1305_encrypt_all_bp(hilda, *args):
    key = hilda.registers.x0.peek(32)
    nonce = hilda.registers.x1.peek(hilda.registers.x2)
    plaintext = hilda.registers.x5
    plaintext_len = hilda.registers.x6
    print(f'_chacha20_poly1305_encrypt_all\n'
          f'    key: {key}\n'
          f'    nonce: {nonce}\n'
          f'    plaintext: {plaintext.peek(plaintext_len)}\n'
          f'---\n')
    hilda.cont()


def _chacha20_poly1305_decrypt_all_bp(hilda, *args):
    key = hilda.registers.x0.peek(32)
    nonce = hilda.registers.x1.peek(hilda.registers.x2)
    encrypted = hilda.registers.x5
    encrypted_len = hilda.registers.x6
    plaintext = hilda.registers.x7
    encrypted_buf = encrypted.peek(encrypted_len)
    hilda.finish()
    print(f'_chacha20_poly1305_decrypt_all\n'
          f'    key: {key}\n'
          f'    nonce: {nonce}\n'
          f'    encrypted: {encrypted_buf}\n'
          f'    plaintext: {plaintext.peek(encrypted_len)}\n'
          f'---\n')
    hilda.cont()


def monitor_crypto_functions() -> None:
    hilda_client = lldb.hilda_client

    hilda_client.symbols.TLV8CopyCoalesced.bp(_TLV8CopyCoalesced_bp)
    hilda_client.symbols.TLV8BufferAppend.bp(_TLV8BufferAppend_bp)
    hilda_client.symbols.SRPClientStart_libsrp.bp(_SRPClientStart_libsrp_bp)
    hilda_client.symbols.SRPClientVerify_libsrp.monitor()
    hilda_client.symbols.SRPServerStart_libsrp.monitor()
    hilda_client.symbols.SRPServerVerify_libsrp.monitor()
    hilda_client.symbols.cced25519_sign.bp(_cced25519_sign_bp)
    hilda_client.symbols.cced25519_verify.bp(_cced25519_verify_bp)
    hilda_client.symbols.CryptoHKDF.bp(_CryptoHKDF_bp)
    hilda_client.symbols._chacha20_poly1305_encrypt_all.bp(_chacha20_poly1305_encrypt_all_bp)
    hilda_client.symbols._chacha20_poly1305_decrypt_all.bp(_chacha20_poly1305_decrypt_all_bp)



#!/usr/bin/xcrun python3

import logging

import coloredlogs
import lldb

from hilda.hilda_client import HildaClient

lldb.hilda_client = None


def hilda(debugger, command, result, internal_dict):
    coloredlogs.install(level=logging.DEBUG)

    if lldb.hilda_client is None:
        lldb.hilda_client = HildaClient(debugger)

    additional_namespace = {'ui': lldb.hilda_client.ui_manager}
    lldb.hilda_client.interact(additional_namespace=additional_namespace)


def __lldb_init_module(debugger, internal_dict):
    debugger.SetAsync(True)
    debugger.HandleCommand('command script add -f lldb_entrypoint.hilda hilda')
    print('''
    👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇👇
    Use `hilda` command to get into Hilda interactive shell.
    If you "continue" before doing so, you cannot do it later.
    👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆👆
    ''')



from hilda.lldb_importer import lldb

_FILENAME = '/tmp/hilda-keylog.txt'


def _ssl_log_secret_bp(hilda, *args):
    label = hilda.registers.x1.peek_str()
    secret = hilda.registers.x2.peek(hilda.registers.x3).hex()
    random = (hilda.registers.x0[6] + 48).peek(32).hex()
    print(f'ssl_log_secret\n'
          f'    label: {label}\n'
          f'    secret: {secret}\n'
          f'    random: {random}\n'
          f'---\n')
    with open(_FILENAME, 'a') as f:
        f.write(f'{label} {random} {secret}\n')
    hilda.cont()


def start_keylog(filename: str = None) -> None:
    global _FILENAME

    if filename is not None:
        _FILENAME = filename
    hilda_client = lldb.hilda_client
    hilda_client.symbols._ZN4bssl14ssl_log_secretEPK6ssl_stPKcNS_4SpanIKhEE.bp(_ssl_log_secret_bp)



import pytest

from hilda.symbol import Symbol


def test_create_symbol_from_int(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    rand = hilda_client.symbols.rand
    assert rand() != Symbol.create(int(rand), hilda_client)()


def test_sanity(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    rand = hilda_client.symbols.rand
    assert rand() != rand()


def test_itemsize(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    addressing_sizes = [8, 4, 2, 1]
    rand = hilda_client.symbols.rand  # type: Symbol
    for i in range(len(addressing_sizes) - 1):
        rand.item_size = addressing_sizes[i]
        long_item = rand[0]
        rand.item_size = addressing_sizes[i + 1]
        short_item = rand[0]
        assert long_item & ((0x100 ** rand.item_size) - 1) == short_item


@pytest.mark.parametrize('augend, addend, result', [
    (1, 1, 2),
    (0, 1, 1),
    (1, 0, 1),
    (1, 1.5, 2.5),
])
def test_symbol_add(hilda_client, augend, addend, result):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    sum_ = hilda_client.symbol(augend) + addend
    reverse_sum = addend + hilda_client.symbol(augend)
    assert reverse_sum == sum_ == result


@pytest.mark.parametrize('minuend, subtrahend, difference', [
    (2, 1, 1),
    (1, 1, 0),
    (1, 0, 1),
])
def test_symbol_sub(hilda_client, minuend, subtrahend, difference):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    diff = hilda_client.symbol(minuend) - subtrahend
    reverse_diff = minuend - hilda_client.symbol(subtrahend)
    assert reverse_diff == diff == difference


@pytest.mark.parametrize('minuend, subtrahend, difference', [
    (2.5, 1, 1.5),
])
def test_symbol_rsub(hilda_client, minuend, subtrahend, difference):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    diff = minuend - hilda_client.symbol(subtrahend)
    assert diff == difference


@pytest.mark.parametrize('multiplier, multiplicand, product', [
    (1, 1, 1),
    (0, 1, 0),
    (1, 2, 2),
    (1, 1.5, 1.5),
])
def test_symbol_mul(hilda_client, multiplier, multiplicand, product):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    prod = hilda_client.symbol(multiplier) * multiplicand
    reverse_prod = multiplicand * hilda_client.symbol(multiplier)
    assert reverse_prod == prod
    assert prod == product


@pytest.mark.parametrize('value, fmt', [
    (1, '0x{:8x}'),
    (0x1122334455667788, '(intptr_t)0x{:x}'),
])
def test_formatting_symbol(hilda_client, value: int, fmt: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param value: Symbol's value.
    :param fmt: How to format symbol.
    """
    symbol = hilda_client.symbol(value)
    assert fmt.format(symbol) == fmt.format(value)


def test_po_cfstr(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert hilda_client.ns('ABC').po() == 'ABC'


def test_po_nsobject(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert hilda_client.objc_get_class('NSObject').new().po().startswith('<NSObject: ')


def test_cf_description_cfstr(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert hilda_client.CFSTR('ABC').cf_description == 'ABC'


def test_cf_description_nsobject(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert hilda_client.objc_get_class('NSObject').new().cf_description.startswith('<NSObject: ')



import shutil
from abc import abstractmethod
from typing import List, Mapping

from click import style
from lldb import SBAddress
from tabulate import tabulate

WORD_SIZE = 8


def dict_diff(d1: Mapping, d2: Mapping) -> List[str]:
    """ Returns a list of keys whose values have changed """
    changed_keys = []
    if d1 is None or d2 is None:
        return changed_keys
    for k, v in d1.items():
        if k in d2 and d2[k] != v:
            changed_keys.append(k)
    return changed_keys


class View:
    """ Base class of view"""

    def __init__(self, hilda_client, title, color_scheme):
        """
        :param hilda_client: hilda.hilda_client.HildaClient
        :param title: str
        :param color_scheme: hilda.ui.ui_manager.ColorScheme
        """
        self.title = title
        self.hilda = hilda_client
        self.active = True
        self.color_scheme = color_scheme

    @abstractmethod
    def __str__(self) -> str:
        return self._format_title()

    def _format_title(self) -> str:
        terminal_size = shutil.get_terminal_size()
        fmt = style(self.title.center(terminal_size.columns, '─'), fg=self.color_scheme.title)
        fmt += '\n'
        return fmt


class StackView(View):
    """ Implements stack view """
    DEFAULT_STACK_DEPTH = 10

    def __init__(self, hilda_client, color_scheme, depth=DEFAULT_STACK_DEPTH):
        """
        :param hilda_client: hilda.hilda_client.HildaClient
        :param depth: int
        :param color_scheme: hilda.ui.ui_manager.ColorScheme

        `prev` saves the last stack stats inorder to perform diff check.
        """
        super().__init__(hilda_client, 'Stack', color_scheme)
        self.depth = depth
        self.prev: Mapping[int, str] = None

    def __str__(self) -> str:
        """ Format stack view for printing """
        stack_mapping = self._create_mapping()
        diff_in_keys = dict_diff(stack_mapping, self.prev)
        self.prev = stack_mapping.copy()
        fmt_parts = []
        offset = 0
        for k, v in stack_mapping.items():
            data_color = self.color_scheme.stack_data if k not in diff_in_keys else self.color_scheme.diff
            fmt = f'{style(hex(k), fg=self.color_scheme.address)}|+' \
                  f'{style(hex(offset), fg=self.color_scheme.stack_offset)}: ' \
                  f'{style(v, fg=data_color)}'
            if offset == 0:
                fmt += f'{"<-- $sp":^50}'
            offset += WORD_SIZE
            fmt_parts.append(fmt)

        return super().__str__() + '\n'.join(fmt_parts)

    def _create_mapping(self) -> Mapping[int, str]:
        """ Generate mapping of stack address:data"""
        base_addr = self.hilda.frame.sp
        stack_mapping = {}
        for i in range(0, self.depth * WORD_SIZE, WORD_SIZE):
            current = base_addr + i
            stack_mapping[current] = '0x' + self.hilda.symbol(current).peek(WORD_SIZE).hex()
        return stack_mapping


class RegistersView(View):
    """ Implements registers view """
    DEFAULT_REGISTERS_TYPE = 'general'

    def __init__(self, hilda_client, color_scheme, rtype=DEFAULT_REGISTERS_TYPE):
        """
        :param hilda_client: hilda.hilda_client.HildaClient
        :param rtype: str
        :param color_scheme: hilda.ui.ui_manager.ColorScheme

        `prev` saves the last registers stats inorder to perform diff check.
        """
        super().__init__(hilda_client, 'Registers', color_scheme)
        self.prev: Mapping[str, str] = None
        self.rtype = rtype

    def __str__(self) -> str:
        """ Format registers view for printing"""
        regs_mapping = self._create_mapping()
        diff_in_keys = dict_diff(regs_mapping, self.prev)
        self.prev = regs_mapping.copy()
        list_of_lists = []
        names = []
        values = []
        i = 0

        # Divide the registers into columns of 16
        for k, v in regs_mapping.items():
            if i % 16 == 0 and i != 0:
                list_of_lists.append(names.copy())
                list_of_lists.append(values.copy())
                names = []
                values = []

            names.append(style(k, fg=self.color_scheme.regs_name))
            if k in diff_in_keys:
                values.append(style(v, fg=self.color_scheme.diff))
            else:
                values.append(style(v, fg=self.color_scheme.regs_value))
            i += 1

        return super().__str__() + tabulate(list(zip(*list_of_lists)), tablefmt='plain', numalign='left')

    def _create_mapping(self) -> Mapping[str, str]:
        """ Generate mapping of registers name:data"""
        regs = self._get_registers()
        regs_mapping = {}
        for reg in regs:
            regs_mapping[reg.GetName()] = reg.GetValue()
        return regs_mapping

    def _get_registers(self):
        """
        Returns list of registers grouped by type.
        Available types: [general,floating]
        :return:  lldb.SBValueList
        """
        registers_set = self.hilda.frame.GetRegisters()
        for value in registers_set:
            if self.rtype.lower() in value.GetName().lower():
                return value
        return None


class DisassemblyView(View):
    """ Implements disassembly view """
    DEFAULT_INSTRUCTION_COUNT = 5
    DEFAULT_DISASSEMBLY_FLAVOR = 'intel'

    def __init__(self, hilda_client, color_scheme, instruction_count=DEFAULT_INSTRUCTION_COUNT,
                 flavor=DEFAULT_DISASSEMBLY_FLAVOR):
        """
        :param hilda_client: hilda.hilda_client.HildaClient
        :param instruction_count: int
        :param color_scheme: hilda.ui.ui_manager.ColorScheme
        """
        super().__init__(hilda_client, 'Disassembly', color_scheme)
        self.instruction_count = instruction_count
        self.flavor = flavor

    def __str__(self) -> str:
        """ Format disassembly view for printing """
        pc = self.hilda.frame.pc
        target = self.hilda.target

        disass = target.ReadInstructions(SBAddress(pc, target), self.instruction_count, self.flavor)
        fmt_parts = []
        for inst in disass:
            load_addr = inst.addr.GetLoadAddress(self.hilda.target)
            file_addr = inst.addr.GetFileAddress()
            base_name = style(inst.addr.module.file.basename, self.color_scheme.basename)
            load_addr = style(hex(load_addr), fg=self.color_scheme.address)
            file_addr = style(hex(file_addr), fg=self.color_scheme.address)
            mnemonic = style(inst.GetMnemonic(self.hilda.target), fg=self.color_scheme.mnemonic)
            operands = style(inst.GetOperands(self.hilda.target), fg=self.color_scheme.operands)
            fmt = f'{base_name}[{load_addr}][{file_addr}]: {mnemonic} {operands}'
            if load_addr == self.hilda.frame.pc:
                fmt += f'{"<-- $pc":^50}'
            fmt_parts.append(fmt)
        return super().__str__() + '\n'.join(fmt_parts)


class BackTraceView(View):
    """ Implements backtrace view """
    DEFAULT_BACKTRACE_DEPTH = 5

    def __init__(self, hilda_client, color_scheme, depth=DEFAULT_BACKTRACE_DEPTH):
        """
        :param hilda_client: hilda.hilda_client.HildaClient
        :param depth: int
        :param color_scheme: hilda.ui.ui_manager.ColorScheme
        """
        super().__init__(hilda_client, 'BackTrace', color_scheme)
        self.depth = depth

    def __str__(self) -> str:
        """ Format backtrace view for printing"""
        bt = self.hilda.bt(should_print=False, depth=self.depth)
        bt_colored = []
        for pair in bt:
            bt_colored.append(
                [
                    style(pair[0], fg=self.color_scheme.address),
                    pair[1]
                ]
            )
        return super().__str__() + tabulate(bt_colored, tablefmt='plain', numalign='left')



import logging
import os
from abc import ABC, abstractmethod
from threading import Thread
from typing import List, Optional

from hilda.exceptions import LLDBException
from hilda.hilda_client import HildaClient
from hilda.lldb_importer import lldb

TIMEOUT = 1
lldb.hilda_client = None

logger = logging.getLogger(__name__)


def hilda(debugger, startup_files: Optional[List[str]] = None):
    if lldb.hilda_client is None:
        lldb.hilda_client = HildaClient(debugger)

    additional_namespace = {'ui': lldb.hilda_client.ui_manager, 'cfg': lldb.hilda_client.configs}
    lldb.hilda_client.interact(additional_namespace=additional_namespace, startup_files=startup_files)


def execute(cmd: str) -> int:
    logging.debug(f'executing: {cmd}')
    return os.system(cmd)


class LLDBListenerThread(Thread, ABC):

    def __init__(self):
        super().__init__()
        lldb.SBDebugger.Initialize()
        self.debugger: lldb.SBDebugger = lldb.SBDebugger.Create()
        self.listener: lldb.SBListener = self.debugger.GetListener()
        self.error: lldb.SBError = lldb.SBError()
        self.debugger.SetAsync(True)
        self.target: lldb.SBTarget = self._create_target()
        self.process: lldb.SBProcess = self._create_process()
        self._check_success()
        self.should_quit = False

    @abstractmethod
    def _create_target(self) -> lldb.SBTarget:
        pass

    @abstractmethod
    def _create_process(self) -> lldb.SBProcess:
        pass

    def _check_success(self):
        if self.error.Success():
            return
        raise LLDBException(self.error.description)

    def run(self):
        event = lldb.SBEvent()
        last_state = lldb.eStateStopped
        while not self.should_quit:
            if not self.listener.WaitForEvent(TIMEOUT, event):
                continue
            if not lldb.SBProcess.EventIsProcessEvent(event):
                continue
            state = self.process.GetStateFromEvent(event)
            if state == lldb.eStateDetached:
                logger.debug('Process Detached')
                self.should_quit = True
            elif state == lldb.eStateExited:
                logger.debug(f'Process Exited with status {self.process.GetExitStatus()}')
                self.should_quit = True
            elif state == lldb.eStateRunning and last_state == lldb.eStateStopped:
                logger.debug("Process Continued")
            elif state == lldb.eStateStopped and last_state == lldb.eStateRunning:
                logger.debug('Process Stopped')
            last_state = state


class LLDBRemote(LLDBListenerThread):
    def __init__(self, address: str, port: int = 1234):
        self.url_connect = f'connect://{address}:{port}'
        super().__init__()

    def _create_target(self) -> lldb.SBTarget:
        return self.debugger.CreateTarget('')

    def _create_process(self) -> lldb.SBProcess:
        logger.debug(f'Connecting to "{self.url_connect}"')
        return self.target.ConnectRemote(self.listener, self.url_connect, None, self.error)


class LLDBAttachPid(LLDBListenerThread):

    def __init__(self, pid: int):
        self.pid = pid
        super().__init__()

    def _create_target(self) -> lldb.SBTarget:
        return self.debugger.CreateTargetWithFileAndArch(None, None)

    def _create_process(self) -> lldb.SBProcess:
        logger.debug(f'Attaching to {self.pid}')
        return self.target.AttachToProcessWithID(self.listener, self.pid, self.error)


class LLDBAttachName(LLDBListenerThread):

    def __init__(self, proc_name: str, wait_for: bool = False):
        self.proc_name = proc_name
        self.wait_for = wait_for
        super().__init__()

    def _create_target(self) -> lldb.SBTarget:
        return self.debugger.CreateTargetWithFileAndArch(None, None)

    def _create_process(self) -> lldb.SBProcess:
        logger.debug(f'Attaching to {self.name}')
        return self.target.AttachToProcessWithName(self.listener, self.proc_name, self.wait_for, self.error)


class LLDBLaunch(LLDBListenerThread):
    def __init__(self, exec_path: str, argv: Optional[List[str]] = None, envp: Optional[List[str]] = None,
                 stdin: Optional[str] = None,
                 stdout: Optional[str] = None, stderr: Optional[str] = None, wd: Optional[str] = None,
                 flags: Optional[int] = 0, stop_at_entry: Optional[bool] = False):
        self.exec_path = exec_path
        self.stdout = stdout
        self.stdin = stdin
        self.stderr = stderr
        self.flags = flags
        self.stop_at_entry = stop_at_entry
        self.argv = argv
        self.envp = envp
        self.working_directory = wd
        super().__init__()

    def _create_target(self) -> lldb.SBTarget:
        return self.debugger.CreateTargetWithFileAndArch(self.exec_path, lldb.LLDB_ARCH_DEFAULT)

    def _create_process(self) -> lldb.SBProcess:
        # Launch(SBTarget self, SBListener listener, char const ** argv, char const ** envp,
        # char const * stdin_path, char const * stdout_path, char const * stderr_path, char const * working_directory,
        # uint32_t launch_flags, bool stop_at_entry, SBError error) -> SBProcess
        logger.debug(f'Lunching process  {self.exec_path}')
        return self.target.Launch(self.listener, self.argv, self.envp,
                                  self.stdin, self.stdout, self.stderr, self.working_directory,
                                  self.flags, self.stop_at_entry,
                                  self.error)


def remote(hostname: str, port: int, startup_files: Optional[List[str]] = None) -> None:
    """ Connect to remote process """
    try:
        lldb_t = LLDBRemote(hostname, port)
        lldb_t.start()
        hilda(lldb_t.debugger, startup_files)
    except LLDBException as e:
        logger.error(e.message)


def attach(name: Optional[str] = None, pid: Optional[int] = None, wait_for: bool = False,
           startup_files: Optional[List[str]] = None) -> None:
    """ Attach to given process and start a lldb shell """
    if (name is not None and pid is not None) or (name is None and pid is None):
        raise ValueError('Provide either a process name or a PID, but not both.')

    try:
        if name is not None:
            lldb_t = LLDBAttachName(name, wait_for)
        else:
            lldb_t = LLDBAttachPid(pid)
        lldb_t.start()
        hilda(lldb_t.debugger, startup_files)
    except LLDBException as e:
        logger.error(e.message)


def launch(exec_path: str, argv: Optional[List] = None, envp: Optional[List] = None,
           stdin: Optional[str] = None,
           stdout: Optional[str] = None, stderr: Optional[str] = None, wd: Optional[str] = None,
           flags: Optional[int] = 0, stop_at_entry: Optional[bool] = False,
           startup_files: Optional[List[str]] = None) -> None:
    """ Launch to given process and start a lldb shell """
    try:
        lldb_t = LLDBLaunch(exec_path, argv, envp, stdin, stdout, stderr, wd, flags, stop_at_entry)
        lldb_t.start()
        hilda(lldb_t.debugger, startup_files)
    except LLDBException as e:
        logger.error(e.message)



import pytest

from hilda.exceptions import SymbolAbsentError


def test_lazy_symbol_getattr(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert hilda_client.symbols.rand()


def test_lazy_symbol_getitem(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert hilda_client.symbols['rand']()


def test_get_absent_symbol(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    with pytest.raises(SymbolAbsentError):
        hilda_client.symbols.symbol_that_doesnt_exist


def test_find(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert 'malloc' not in hilda_client.symbols, 'expected malloc to not be present'

    # populate the jar
    hilda_client.symbols.malloc
    hilda_client.symbols.rand

    assert 1 == len(hilda_client.symbols.clean().find('malloc')), 'expected to find only one'
